unit Element;

interface

uses Windows, Gauge;

type
  TElement = class(TObject)
    X, Y: Integer;
    Width: Integer;
    Height: Integer;
    R, G, B: Integer;
    FontName: string;
    FontSize: Integer;
    FontBold: Boolean;
    FontItalic: Boolean;
    Gauges: TGauges;
    constructor Create(AX, AY, AWidth, AHeight, AR, AG, AB: Integer;
      AFontName: string; AFontSize: Integer; AFontBold, AFontItalic: Boolean;
      const Args: array of string; const AGauges: TGauges);
    procedure Init(const Args: array of string); virtual; abstract;
    procedure Draw(DC: HDC); virtual; abstract;
    procedure Update; virtual; abstract;
  end;
  TElementClass = class of TElement;

  PPCharArray = ^PCharArray;
  PCharArray = array[0..Maxint div SizeOf(PChar)-1] of PChar;
  TPlugin = class(TElement)
  private
    Handle: THandle;
    Id: Integer;
    InitProc: function (X, Y, Width, Height, R, G, B: Integer;
      ArgCount: Integer; Args: PPCharArray): Integer; stdcall;
    DrawProc: procedure (Id: Integer; HDC: HDC;
    X, Y, Width, Height, R, G, B: Integer); stdcall;
    UpdateProc: procedure (Id: Integer); stdcall;
    UnloadProc: procedure (Id: Integer); stdcall;
  public   
    destructor Destroy; override;
    procedure Init(const Args: array of string); override;
    procedure Draw(DC: HDC); override;
    procedure Update; override;
  end;

  TText = class(TElement)
  private
    Text: string;
    LText: string;
    RText: string;
    Fmt: string;
    Args: array[0..15] of string;
  public
    procedure Init(const Args: array of string); override;
    procedure Draw(DC: HDC); override;
    procedure Update; override;
  end;

  TScrollText = class(TText)
  private
    Left: Integer;
  public
    procedure Draw(DC: HDC); override;
    procedure Update; override;
  end;

  TBarGraph = class(TElement)
  private
    Len: Integer;
    Data: string;
    DataMax: string;
  public
    procedure Init(const Args: array of string); override;
    procedure Draw(DC: HDC); override;
    procedure Update; override;
  end;

  PIntArray = ^IntArray;
  IntArray = array[0..MaxInt div SizeOf(Integer)-1] of Integer;
  TAreaGraph = class(TElement)
  private
    DataLog: PIntArray;
    DataInd: Integer;
    Data: string;
    DataMax: string;
  public
    destructor Destroy; override;
    procedure Init(const Args: array of string); override;
    procedure Draw(DC: HDC); override;
    procedure Update; override;
  end;
  
  TLineGraph = class(TAreaGraph)
    procedure Draw(DC: HDC); override;
  end;

  TDClock = class(TElement)
  private
    Text0, Text1: string;
    Beat: Char;
    Hour, Min, Sec: string;
    procedure DrawChar(DC: HDC; X, Y, W, H: Integer; C: Char; Brush: HBRUSH);
  public
    procedure Init(const Args: array of string); override;
    procedure Draw(DC: HDC); override;
    procedure Update; override;
  end;

  TMoon = class(TElement)
  private
    Phase: string;
    PrevPhase: Double;
    Rgn: HRGN;
    Bmp: HBITMAP;
    BmpWidth, BmpHeight: Integer;
    function CreateRgn(Left, Top, Width, Height: Integer; Phase: Double): HRGN;
  public
    procedure Init(const Args: array of string); override;
    procedure Draw(DC: HDC); override;
    procedure Update; override;
  end;

  TAClock = class(TElement)
  private
    HHand, MHand, SHand: Integer;
    Hour, Min, Sec: string;
  public
    procedure Init(const Args: array of string); override;
    procedure Draw(DC: HDC); override;
    procedure Update; override;
  end;

  TBClock = class(TElement)
  private
    Value: array[0..5] of Integer;
    Hour, Min, Sec: string;
  public
    procedure Init(const Args: array of string); override;
    procedure Draw(DC: HDC); override;
    procedure Update; override;
  end;

var
  ElementClasses: array[0..9] of TElementClass = (
    TPlugin, TText, TScrollText, TBarGraph, TAreaGraph, TLineGraph, TDClock,
    TMoon, TAClock, TBClock
  );

implementation

uses NLib;

procedure RectFill(DC: HDC; X, Y, W, H: Integer; Brush: HBRUSH);
var
  r: TRect;
begin
  r.Left := X;
  r.Top := Y;
  r.Right := X+W;
  r.Bottom := Y+H;
  FillRect(DC, r, Brush);
end;

constructor TElement.Create(AX, AY, AWidth, AHeight, AR, AG, AB: Integer;
  AFontName: string; AFontSize: Integer; AFontBold, AFontItalic: Boolean;
  const Args: array of string; const AGauges: TGauges);
begin
  X := AX;
  Y := AY;
  Width := AWidth;
  Height := AHeight;
  R := AR;
  G := AG;
  B := AB;
  FontName := AFontName;
  FontSize := AFontSize;
  FontBold := AFontBold;
  FontItalic := AFontItalic;
  Gauges := AGauges;
  Init(Args);
end;

{ plugin }

destructor TPlugin.Destroy;
begin
  if Assigned(UnloadProc) then
    UnloadProc(Id);
  if Handle <> 0 then  
    FreeLibrary(Handle);
  inherited;
end;

procedure TPlugin.Init(const Args: array of string);
var
  i: Integer;
  pargs: PPCharArray;
begin
  Handle := LoadLibrary(PChar(Args[2]+'.nmp'));
  if Handle = 0 then
    Gauges.Log(Self, Args[2] + ': can''t load')
  else begin
    Gauges.Log(Self, Args[2] + ': loaded');
    InitProc := GetProcAddress(Handle, 'init');
    DrawProc := GetProcAddress(Handle, 'draw');
    UpdateProc := GetProcAddress(Handle, 'update');
    UnloadProc := GetProcAddress(Handle, 'unload');
    if Assigned(InitProc) then
    begin
      GetMem(pargs, SizeOf(PChar)*(High(Args)+1));
      for i := 0 to High(Args) do
	pargs[i] := PChar(Args[i]);
      Id := InitProc(X, Y, Width, Height, R, G, B, 0, pargs);
      FreeMem(pargs);
    end;
  end;
end;

procedure TPlugin.Draw(DC: HDC);
begin
  if Assigned(DrawProc) then
    DrawProc(Id, DC, X, Y, Width, Height, R, G, B);
end;

procedure TPlugin.Update;
begin
  if Assigned(UpdateProc) then
    UpdateProc(Id);
end;

{ text }

procedure TText.Init(const Args: array of string);
var
  i: Integer;
begin
  for i := 2 to High(Args) do
    case i of
      2:  Fmt := Args[i];
    else
      Self.Args[i-3] := Args[i];
    end;
end;

procedure TText.Draw(DC: HDC);
var
  rc: TRect;
begin
  SetTextColor(DC, RGB(R, G, B));
  rc.Left := X;
  rc.Top := Y;
  rc.Right := Width;
  rc.Bottom := rc.Top+Height;
  if LText <> '' then
    DrawText(DC, PChar(LText), Length(LText), rc,
      DT_NOPREFIX or DT_NOCLIP or DT_LEFT);
  if Text <> '' then
    DrawText(DC, PChar(Text), Length(Text), rc,
      DT_NOPREFIX or DT_NOCLIP or DT_CENTER);
  if RText <> '' then
    DrawText(DC, PChar(RText), Length(RText), rc,
      DT_NOPREFIX or DT_NOCLIP or DT_RIGHT);
end;

procedure TText.Update;
var
  i, ac: Integer;
  tab: Boolean;
begin
  Text := '';
  LText := '';
  RText := '';
  ac := 0;
  i := 0;
  tab := False;
  while i < Length(Fmt) do
  begin
    Inc(i);
    if Fmt[i] = '%' then
    begin
      Inc(i);
      if i <= Length(Fmt) then
        case Fmt[i] of
          's': begin
                 Text := Text + Gauges[Args[ac]];
                 Inc(ac);
               end;
          'h': begin
                 Text := Text + UnitFormat(SInt64(Gauges[Args[ac]]));
                 Inc(ac);
               end;
          '%': Text := Text + '%';
        end;
    end
    else if Fmt[i] = '\' then
    begin
      Inc(i);
      if i <= Length(Fmt) then
        case Fmt[i] of
          't': begin
                 tab := True;
                 LText := Text;
                 Text := '';
               end;
          '\': Text := Text + '\';
        end;
    end else
      Text := Text + Fmt[i];
  end;
  if tab then
  begin
    RText := Text;
    Text := '';
  end;
end;

{ scrolltext }

procedure TScrollText.Draw(DC: HDC);
var
  rc: TRect;
  size: TSize;
  s: string;
begin
  GetTextExtentPoint32(DC, PChar(Text), Length(Text), size);
  SetTextColor(DC, RGB(R, G, B));
  if size.cx > Width then
  begin
    s := Text + '  ';
    GetTextExtentPoint32(DC, PChar(s), Length(s), size);
    rc.Left := X + Left;
    rc.Top := Y;
    rc.Right := Width;
    rc.Bottom := rc.Top+Height;
    s := s + Text;
    DrawText(DC, PChar(s), Length(s), rc, DT_NOPREFIX or DT_NOCLIP or DT_LEFT);
    Dec(Left, 2);
    if Left < -size.cx then
      Left := 0;
  end
  else begin
    rc.Left := X;
    rc.Top := Y;
    rc.Right := Width;
    rc.Bottom := rc.Top+Height;
    DrawText(DC, PChar(Text), Length(Text), rc,
      DT_NOPREFIX or DT_NOCLIP or DT_LEFT);
  end;
end;

procedure TScrollText.Update;
var
  prevtext: string;
begin
  prevtext := Text;
  inherited;
  if Text <> prevtext then
    Left := 0;
end;

{ bargraph }

procedure TBarGraph.Init(const Args: array of string);
var
  i: Integer;
begin
  for i := 2 to High(Args) do
    case i of
      2: Data := Args[i];
      3: DataMax := Args[i];
    end;
end;

procedure TBarGraph.Draw(DC: HDC);
var
  i: Integer;
  br: HBRUSH;
begin
  br := CreateSolidBrush(RGB(R, G, B));
  for i := 0 to (Len div 3)-1 do
    RectFill(DC, X+i*3, Y, 2, Height, br);
  DeleteObject(br);
end;

procedure TBarGraph.Update;
begin
  if SFloat(Gauges[DataMax]) = 0 then
    Len := 0
  else
    Len := Trunc(SFloat(Gauges[Data]) * Width / SFloat(Gauges[DataMax]));
end;

{ areagraph }

procedure TAreaGraph.Init(const Args: array of string);
var
  i: Integer;
begin
  for i := 2 to High(Args) do
    case i of
      2: Data := Args[i];
      3: DataMax := Args[i];
    end;
end;

destructor TAreaGraph.Destroy;
begin
  FreeMem(DataLog);
end;

procedure TAreaGraph.Draw(DC: HDC);
var
  i, d: Integer;
  br: HBRUSH;
begin
  br := CreateSolidBrush(RGB(R, G, B));
  for i := 0 to Width-1 do
  begin
    d := DataLog[(i+DataInd) mod Width];
    RectFill(DC, X+i, Y+Height-d, 1, d, br);
  end;
  DeleteObject(br);
end;

procedure TAreaGraph.Update;
begin
  if not Assigned(DataLog) then
  begin
    GetMem(DataLog, Width*SizeOf(Integer));
    FillChar(DataLog^, Width*SizeOf(Integer), 0);
  end;
  if SFloat(Gauges[DataMax]) = 0 then
    DataLog[DataInd] := 0
  else begin
    DataLog[DataInd] :=
      Trunc(SFloat(Gauges[Data]) * (Height{-1}) / SFloat(Gauges[DataMax]));
    if DataLog[DataInd] > Height then
      DataLog[DataInd] := Height
    else if DataLog[DataInd] < 0 then
      DataLog[DataInd] := 0;
  end;
  Inc(DataInd);
  if DataInd = Width then
    DataInd := 0;
end;

{ linegraph }

procedure TLineGraph.Draw(DC: HDC);
var
  i, d, prevd: Integer;
  pen, oldpen: HPEN;
begin
  pen := CreatePen(PS_SOLID, 0, RGB(R, G, B));
  oldpen := SelectObject(DC, pen);
  MoveToEx(DC, X, Y+Height-DataLog[DataInd mod Width], nil);
  prevd := 0;
  for i := 0 to Width-1 do
  begin
    d := DataLog[(i+DataInd) mod Width];
    if (d = 0) and (prevd = 0) then
      MoveToEx(DC, X+i, Y+Height-d-1, nil)
    else
      LineTo(DC, X+i, Y+Height-d);
    prevd := d;
  end;
  SelectObject(DC, oldpen);
  DeleteObject(pen);
end;

{ dclock }

procedure TDClock.Init(const Args: array of string);
var
  i: Integer;
begin
  for i := 2 to High(Args) do
    case i of
      2: Hour := Args[i];
      3: Min := Args[i];
      4: Sec := Args[i];
    end;
end;

procedure TDClock.DrawChar(DC: HDC; X, Y, W, H: Integer; C: Char;
  Brush: HBRUSH);
const
  {
     00
    3  4
     11
    5  6
     22
  }
  tab: array[0..9, 0..6] of Integer = (
    (1,0,1,1,1,1,1),
    (0,0,0,0,1,0,1),
    (1,1,1,0,1,1,0),
    (1,1,1,0,1,0,1),
    (0,1,0,1,1,0,1),
    (1,1,1,1,0,0,1),
    (1,1,1,1,0,1,1),
    (1,0,0,0,1,0,1),
    (1,1,1,1,1,1,1),
    (1,1,1,1,1,0,1)
  );
var
  n: Integer;
begin
  case c of
    ':': RectFill(DC, x, y+h div 2, 2, 2, Brush); // upper dot
    '.': RectFill(DC, x, y+h*2+1-h div 2, 2, 2, Brush); // lower dot
    '0'..'9':
      begin
        n := Ord(c) - Ord('0');
        if tab[n,0] = 1 then RectFill(DC, x+1,   y,       w-2, 1, Brush);
        if tab[n,1] = 1 then RectFill(DC, x+1,   y+h+1,   w-2, 1, Brush);
        if tab[n,2] = 1 then RectFill(DC, x+1,   y+h*2+2, w-2, 1, Brush);
        if tab[n,3] = 1 then RectFill(DC, x,     y+1,     2,   h, Brush);
        if tab[n,4] = 1 then RectFill(DC, x+w-2, y+1,     2,   h, Brush);
        if tab[n,5] = 1 then RectFill(DC, x,     y+h+2,   2,   h, Brush);
        if tab[n,6] = 1 then RectFill(DC, x+w-2, y+h+2,   2,   h, Brush);
      end;
  end;
end;

procedure TDClock.Draw(DC: HDC);
const
  HLineHeight = 1;
  dotw = 2;
var
  x, w, h, sp, tw: Integer;
  br: HBRUSH;
begin
  w := Trunc(Width / 5 + 0.5); // "00:00" length is 5
  sp := Trunc(w / 6 + 0.5); // letter space
  w := w - sp; // digit width
  tw := w*4 + dotw+2 + sp*4; // total width
  h := (Height - HLineHeight*3) div 2;
  br := CreateSolidBrush(RGB(R, G, B));
  x := Self.X + (Width - tw) div 2;
  DrawChar(DC, x, Y, w, h, Text0[1], br);
  Inc(x, w+sp);
  DrawChar(DC, x, Y, w, h, Text0[2], br);
  Inc(x, w+sp+1);
  DrawChar(DC, x, Y, w, h, Beat, br);
  Inc(x, dotw+1+sp);
  DrawChar(DC, x, Y, w, h, Text1[1], br);
  Inc(x, w+sp);
  DrawChar(DC, x, Y, w, h, Text1[2], br);
  DeleteObject(br);
end;

procedure TDClock.Update;
begin
  Text0 := Gauges[Hour] + '  '; // minmum length is 2
  Text1 := Gauges[Min] + '  '; // minmum length is 2
  if SInt(Gauges[Sec]) mod 2 = 0 then
    Beat := ':'
  else
    Beat := '.';
end;

{ moon }

procedure TMoon.Init(const Args: array of string);
var
  b: TBitmap;
begin
  if Length(Args) > 2 then
    Phase := Args[2]
  else
    Phase := '';
  Bmp := LoadBitmap(HInstance, 'MOONBITMAP');
  GetObject(Bmp, SizeOf(b), @b);
  BmpWidth := b.bmWidth;
  BmpHeight := b.bmHeight;
end;

procedure TMoon.Draw(DC: HDC);
var
  //br: HBRUSH;
  oldbmp: HBITMAP;
  mdc: HDC;
begin
  //br := CreateSolidBrush(RGB(R, G, B));
  //FillRgn(DC, Rgn, br);
  //DeleteObject(br);
  if Rgn = 0 then
    Exit;
  mdc := CreateCompatibleDC(DC);
  oldbmp := SelectObject(mdc, Bmp);
  SelectClipRgn(DC, Rgn);
  SetStretchBltMode(DC, COLORONCOLOR);
  StretchBlt(DC, X, Y, Height, Height, mdc, 0, 0, BmpWidth, BmpHeight,
    SRCCOPY);
  SelectClipRgn(DC, 0);
  SelectObject(mdc, oldbmp);
  DeleteDC(mdc);
end;

function TMoon.CreateRgn(Left, Top, Width, Height: Integer;
  Phase: Double): HRGN;
var
  rgn: HRGN;
  n: Integer;
begin
  Result := CreateEllipticRgn(Left, Top, Left+Width, Top+Height); // circle
  if Phase > 14 then // right mask
    rgn := CreateRectRgn(Left+Width div 2, Top, Left+Width, Top+Height)
  else // left mask
    rgn := CreateRectRgn(Left, Top, Left+Width div 2, Top+Height); 
  n := Trunc(Width * (1.0 - Abs(Cos(Pi*2*Phase/28.0)))/2 + 0.5);
  CombineRgn(Result, Result, rgn, RGN_DIFF);
  DeleteObject(rgn);
  rgn := CreateEllipticRgn(Left+n, Top, Left+Width-n, Top+Height);
  if (Phase > 7) and (Phase < 21) then
    CombineRgn(Result, Result, rgn, RGN_OR)
  else
    CombineRgn(Result, Result, rgn, RGN_DIFF);
  DeleteObject(rgn);
end;

procedure TMoon.Update;
var
  p: Double;
begin
  p := SFloat(Gauges[Phase]);
  if p = PrevPhase then
    Exit;
  if Rgn <> 0 then
    DeleteObject(Rgn);
  Rgn := CreateRgn(X, Y, Height, Height, p);
  PrevPhase := p;
end;

{ aclock }

procedure TAClock.Init(const Args: array of string);
var
  i: Integer;
begin
  for i := 2 to High(Args) do
    case i of
      2: Hour := Args[i];
      3: Min := Args[i];
      4: Sec := Args[i];
    end;
end;

procedure TAClock.Draw(DC: HDC);
var
  pen, oldpen: HPEN;
  i, rx, ry: Integer;
begin
  pen := CreatePen(PS_SOLID, 0, RGB(R, G, B));
  oldpen := SelectObject(DC, pen);
  //Arc(DC, X, Y, X+Width, Y+Height, 0, 0, 0, 0);
  rx := Width div 2;
  ry := Height div 2;
  for i := 0 to 11 do
    //SetPixelV(DC, X+rx+Trunc(Cos(2*pi*i/12.0)*(rx*0.9)),
    //  Y+ry+Trunc(Sin(2*pi*i/12.0)*(ry*0.9)), RGB(R, G, B));
    SetPixelV(DC, X+rx+Trunc(Cos(2*pi*i/12.0)*rx),
      Y+ry+Trunc(Sin(2*pi*i/12.0)*ry), RGB(R, G, B));
  if Sec <> '' then
  begin
    MoveToEx(DC, X+rx, Y+ry, nil);
    LineTo(DC,
      X+rx+Trunc(Cos(2*pi*SHand/60.0-pi/2)*(rx*0.9)),
      Y+ry+Trunc(Sin(2*pi*SHand/60.0-pi/2)*(ry*0.9)));
  end;
  SelectObject(DC, oldpen);
  DeleteObject(pen);
  pen := CreatePen(PS_SOLID, 2, RGB(R, G, B));
  oldpen := SelectObject(DC, pen);
  MoveToEx(DC, X+rx, Y+ry, nil);
  LineTo(DC,
    X+rx+Trunc(Cos(2*pi*MHand/60.0-pi/2)*(rx*0.9)),
    Y+ry+Trunc(Sin(2*pi*MHand/60.0-pi/2)*(ry*0.9)));
  MoveToEx(DC, X+rx, Y+ry, nil);
  LineTo(DC,
    X+rx+Trunc(Cos(2*pi*(HHand+MHand/60)/12.0-pi/2)*(rx*0.7)),
    Y+ry+Trunc(Sin(2*pi*(HHand+MHand/60)/12.0-pi/2)*(ry*0.7)));
  SelectObject(DC, oldpen);
  DeleteObject(pen);
end;

procedure TAClock.Update;
begin
  HHand := SInt(Gauges[Hour]);
  MHand := SInt(Gauges[Min]);
  SHand := SInt(Gauges[Sec]);
end;

{ bclock }

procedure TBClock.Init(const Args: array of string);
var
  i: Integer;
begin
  for i := 2 to High(Args) do
    case i of
      2: Hour := Args[i];
      3: Min := Args[i];
      4: Sec := Args[i];
    end;
end;

procedure TBClock.Draw(DC: HDC);
var
  i, xx, yy, w, h, v: Integer;
  br: HBRUSH;
begin
  br := CreateSolidBrush(RGB(R, G, B));
  w := (Width - 5) div 6;
  h := (Height - 3) div 4;
  xx := X;
  for i := 0 to 5 do
  begin
    v := Value[i];
    yy := Y + Height - h;
    while v > 0 do
    begin
      if v and 1 = 1 then
        RectFill(DC, xx, yy, w, h, br);
      v := v shr 1;
      Dec(yy, h+1);
    end;
    inc(xx, w+1);
  end;
  DeleteObject(br);
end;

procedure TBClock.Update;
var
  h, m, s: Integer;
begin
  h := SInt(Gauges[Hour]);
  m := SInt(Gauges[Min]);
  s := SInt(Gauges[Sec]);
  Value[0] := h div 10;
  Value[1] := h mod 10;
  Value[2] := m div 10;
  Value[3] := m mod 10;
  Value[4] := s div 10;
  Value[5] := s mod 10;
end;

end.
