unit Layout;

interface

uses Windows, NUtil, Element, Gauge;

type
  TLayouter = class(TObject)
    X, Width, Height: Integer;
    R, G, B: Integer;
    BgR, BgG, BgB, BgA: Integer;
    FontName: string;
    FontSize: Integer;
    FontBold: Boolean;
    FontItalic: Boolean;
    Elements: TObjList;
    constructor Create;
    destructor Destroy; override;
    procedure Draw(DC: HDC);
    procedure Update;
    procedure Eval(const S: string; const Gauges: TGauges);
    procedure Load(const Path: string; const Gauges: TGauges);
    function NameToClass(const Name: string): TElementClass;
  end;

implementation

uses Types, NLib, NConf, LayoutDefault;

constructor TLayouter.Create;
begin
  Elements := TObjList.Create;
end;

destructor TLayouter.Destroy;
begin
  Elements.Free;
  inherited;
end;

procedure TLayouter.Draw(DC: HDC);
var
  i: Integer;
  font, oldfont: HFONT;
  fn: string;
  fs: Integer;
  fb: Boolean;
  fi: Boolean;
  fw: DWORD;
begin
  fn := '';
  fs := 0;
  fb := False;
  fi := False;
  font := 0;
  oldfont := 0;
  SetBkMode(DC, TRANSPARENT);
  for i := 0 to Elements.Count-1 do
    with Elements[i] as TElement do
    begin
      if (FontName <> fn) or (FontSize <> fs) or (FontBold <> fb) or
        (FontItalic <> fi) then
      begin
        if FontBold then fw := FW_BOLD else fw := FW_NORMAL;
        if (font <> 0) and (oldfont <> 0) then
        begin
          SelectObject(DC, oldfont);
          DeleteObject(font);
        end;
        font := CreateFont(FontSize, 0, 0, 0, fw, Ord(FontItalic), 0, 0,
          DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
          DEFAULT_QUALITY, DEFAULT_PITCH, PChar(FontName));
        oldfont := SelectObject(DC, font);
      end;
      Draw(DC);
    end;
  SelectObject(DC, oldfont);
  DeleteObject(font);
end;

procedure TLayouter.Update;
var
  i: Integer;
begin
  for i := 0 to Elements.Count-1 do
    (Elements[i] as TElement).Update;
end;

function TLayouter.NameToClass(const Name: string): TElementClass;
var
  i: Integer;
begin
  for i := 0 to High(ElementClasses) do
    if ElementClasses[i].ClassNameIs('T'+Name) then
    begin
      Result := ElementClasses[i];
      Exit;
    end;
  Result := nil;
end;

procedure TLayouter.Eval(const S: string; const Gauges: TGauges);
var
  args: TStringDynArray;
  cmd: string;
  c: TElementClass;
  i: Integer;
begin
  args := SplitCommandLine(S);
  cmd := args[0];
  if cmd = 'width' then
    for i := 1 to High(args) do
      case i of
        1: Width := SInt(args[i]);
      end
  else if cmd = 'bgcolor' then
    for i := 1 to High(args) do
      case i of
        1: BgR := SInt(args[i]);
        2: BgG := SInt(args[i]);
        3: BgB := SInt(args[i]);
      end
  else if cmd = 'bgalpha' then
    for i := 1 to High(args) do
      case i of
        1: BgA := SInt(args[i]);
      end
  else if cmd = 'space' then
    for i := 1 to High(args) do
      case i of
        1: Inc(Height, SInt(args[i]));
      end
  else if cmd = 'color' then
    for i := 1 to High(args) do
      case i of
        1: R := SInt(args[i]);
        2: G := SInt(args[i]);
        3: B := SInt(args[i]);
      end
  else if cmd = 'font' then
    for i := 1 to High(args) do
      case i of
        1: FontName := args[i];
        2: FontSize := SInt(args[i]);
        3, 4:
          begin
            FontBold := FontBold or (args[i] = 'bold');
            FontItalic := FontItalic or (args[i] = 'italic');
          end;
      end
  else begin
    c := NameToClass(cmd);
    if Assigned(c) and (Length(args) > 1) then
    begin
      Elements.Add(c.Create(X, Height, Width-X, SInt(args[1]), r, g, b,
        FontName, FontSize, FontBold, FontItalic, args, Gauges));
      Inc(Height, SInt(args[1]));
    end;
  end;
end;

procedure TLayouter.Load(const Path: string; const Gauges: TGauges);
var
  i: Integer;
begin
  Elements.Clear;
  // Default attributes
  X := 1; // left margin
  Width := 48;
  Height := 0;
  R := 255;
  G := 255;
  B := 255;
  BgA := 128;
  FontName := 'Arial';
  FontSize := 14;
  FontBold := False;
  FontItalic := False;
  try
    with TConf.Create(Path) do
      try
        while ReadLine do
          Eval(Line, Gauges);
      finally
        Free;
      end;
  except
    on E: EIOError do // use default layout
      for i := 0 to High(LayoutDefault.Layout) do
        Eval(LayoutDefault.Layout[i], Gauges);
  end;
end;

end.
